package net.sf.brunneng.fusie.core.expressions

import net.sf.brunneng.fusie.core.Constants

/**
 * User: goodg_000
 * Date: 30.03.13
 * Time: 23:36
 */
object F2ExpressionFactory {

  // Numeric operations ################################################################################################
  def plus[T](expr1: Expression[T], expr2: Expression[T])
             (implicit numeric: Numeric[T]): Expression[T] = {
    new F2Expression[T, T, T](expr1, expr2, (v1, v2) => numeric.plus(v1, v2), "+")
  }

  def minus[T](expr1: Expression[T], expr2: Expression[T])
              (implicit numeric: Numeric[T]): Expression[T] = {
    new F2Expression[T, T, T](expr1, expr2, (v1, v2) => numeric.minus(v1, v2), "-")
  }

  def multiply[T](expr1: Expression[T], expr2: Expression[T])
                 (implicit numeric: Numeric[T]): Expression[T] = {
    new F2Expression[T, T, T](expr1, expr2, (v1, v2) => numeric.times(v1, v2), "*")
  }

  def div[T](expr1: Expression[T], expr2: Expression[T])
            (implicit fractional: Fractional[T]): Expression[T] = {
    new F2Expression[T, T, T](expr1, expr2, (v1, v2) => fractional.div(v1, v2), "/")
  }

  def integralDiv[T](expr1: Expression[T], expr2: Expression[T])
            (implicit integral: Integral[T]): Expression[T] = {
    new F2Expression[T, T, T](expr1, expr2, (v1, v2) => integral.quot(v1, v2), "/")
  }

  def pow[T](expr1: Expression[T], expr2: Expression[T])
            (implicit numeric: Numeric[T]): Expression[Double] = {
    new F2Expression[T, T, Double](expr1, expr2,
      (v1, v2) => scala.math.pow(numeric.toDouble(v1), numeric.toDouble(v2)), "^")
  }

  // Comparison operations #############################################################################################

  def eq[T](expr1: Expression[T], expr2: Expression[T])
  : Expression[Boolean] = {
    new F2Expression[T, T, Boolean](expr1, expr2, (v1, v2) => v1 == v2, "==")
  }

  def eqDouble(expr1: Expression[Double], expr2: Expression[Double])
  : Expression[Boolean] = {
    new F2Expression[Double, Double, Boolean](expr1, expr2, (v1, v2) => scala.math.abs(v1 - v2) < Constants.EPSILON, "==")
  }

  def notEq[T](expr1: Expression[T], expr2: Expression[T])
  : Expression[Boolean] = {
    new F2Expression[T, T, Boolean](expr1, expr2, (v1, v2) => v1 != v2, "!=")
  }

  def notEqDouble(expr1: Expression[Double], expr2: Expression[Double])
  : Expression[Boolean] = {
    new F2Expression[Double, Double, Boolean](expr1, expr2, (v1, v2) => scala.math.abs(v1 - v2) >= Constants.EPSILON, "!=")
  }

  def lt[T](expr1: Expression[T], expr2: Expression[T])
           (implicit numeric: Numeric[T]): Expression[Boolean] = {
    new F2Expression[T, T, Boolean](expr1, expr2, (v1, v2) => numeric.lt(v1, v2), "<")
  }

  def gt[T](expr1: Expression[T], expr2: Expression[T])
           (implicit numeric: Numeric[T]): Expression[Boolean] = {
    new F2Expression[T, T, Boolean](expr1, expr2, (v1, v2) => numeric.gt(v1, v2), ">")
  }

  def lteq[T](expr1: Expression[T], expr2: Expression[T])
             (implicit numeric: Numeric[T]): Expression[Boolean] = {
    new F2Expression[T, T, Boolean](expr1, expr2, (v1, v2) => numeric.lteq(v1, v2), "<=")
  }

  def gteq[T](expr1: Expression[T], expr2: Expression[T])
             (implicit numeric: Numeric[T]): Expression[Boolean] = {
    new F2Expression[T, T, Boolean](expr1, expr2, (v1, v2) => numeric.gteq(v1, v2), ">=")
  }

  // Boolean operations ################################################################################################
  def and(expr1: Expression[Boolean], expr2: Expression[Boolean])
  : Expression[Boolean] = {
    new F2Expression[Boolean, Boolean, Boolean](expr1, expr2, (v1, v2) => v1 && v2, "&&")
  }

  def or(expr1: Expression[Boolean], expr2: Expression[Boolean])
  : Expression[Boolean] = {
    new F2Expression[Boolean, Boolean, Boolean](expr1, expr2, (v1, v2) => v1 || v2, "||")
  }

  def andNullSafe(expr1: Expression[Boolean], expr2: Expression[Boolean])
  : Expression[Boolean] = {
    if (expr1 == null && expr2 != null)
      expr2
    else if (expr1 != null && expr2 == null)
      expr1
    else if (expr1 == null && expr2 == null)
      null
    else
      new F2Expression[Boolean, Boolean, Boolean](expr1, expr2, (v1, v2) => v1 && v2, "&&")
  }

  def orNullSafe(expr1: Expression[Boolean], expr2: Expression[Boolean])
  : Expression[Boolean] = {
    if (expr1 == null && expr2 != null)
      expr2
    else if (expr1 != null && expr2 == null)
      expr1
    else if (expr1 == null && expr2 == null)
      null
    else
      new F2Expression[Boolean, Boolean, Boolean](expr1, expr2, (v1, v2) => v1 || v2, "||")
  }
}
