package net.sf.brunneng.fusie.engine.fcf

import net.sf.brunneng.fusie.core._
import scala.collection.mutable
import net.sf.brunneng.fusie.engine.{InferenceEngineInput, InferenceEngine}
import scala.collection.mutable.ListBuffer
import java.util
import net.sf.brunneng.fusie.core.expressions.fuzzy.FuzzyBooleanExpression

/**
 * Forward Chaining Fuzzy logic inference engine.<br>
 * Fuzzy logic means that in case of rule conflict, resolution of it assumes that both rules can be executed, and
 * split inference on N subtrees, where N is number of rules in conflicted group. Then after finishing inference
 * in all subtrees it able to calculate probability of every possible inferred value of target variable.
 *
 * User: goodg_000
 * Date: 08.05.13
 * Time: 21:52
 */
class FCFInferenceEngine extends InferenceEngine {

  class FCNode {
    private var context: Context = null
    private var probabilityContext: Context = null
    private var actualRules = new mutable.HashSet[Rule]
    private var successRules = new mutable.HashSet[Rule]
    private var failedRules = new mutable.ListBuffer[Rule]
    private var cantBeExecutedRules = new mutable.ListBuffer[Rule]

    private var parent: FCNode = null
    private var children = new mutable.ListBuffer[FCNode]

    def this(parent: FCNode, context: Context, probabilityContext: Context, actualRules: List[Rule]) {
      this()
      this.parent = parent
      this.context = context
      this.probabilityContext = probabilityContext
      allContexts += context

      this.actualRules ++= actualRules
    }

    def populate() {
      env.pushContext(context)
      populateInternal()
      env.popContext()
    }

    private def populateInternal() {
      allFuzzyExpressions.foreach(e => e.probabilityContext = probabilityContext)

      var repeat = true
      val exceptRules = getRulesInGroups(getRulesGroups)
      while (repeat) {
        var repeatExecuteSimpleRules = true

        var simpleReadyRules: List[Rule] = null
        while (repeatExecuteSimpleRules) {
          if (simpleReadyRules == null) {
            simpleReadyRules = actualRules.filter(r => r.isReadyForExecution && !exceptRules.contains(r)).toList
          }
          else {
            val candidates = simpleReadyRules.flatMap(
              r => r.conclusions.map(c => c.variable.varName)).distinct.flatMap(
                name => requiredVariableNameToRulesMap.getOrElse(name, new mutable.ListBuffer[Rule])).distinct
            simpleReadyRules = candidates.filter(
              r => actualRules.contains(r) && r.isReadyForExecution && !exceptRules.contains(r)).toList
          }

          simpleReadyRules.foreach(executeRule(_))
          if (simpleReadyRules.isEmpty) {
            repeatExecuteSimpleRules = false
          }
        }

        if (env.contains(input.goalVariableName)) {
          return
        }

        val readyGroups = getRulesGroups.filter(e => e._2.forall(r => r.isReadyForExecution))
        val rulesInGroups = getRulesInGroups(readyGroups).toList
        val successRulesInGroups = rulesInGroups.filter(r => r.allPreconditionsAreTrue)
        if (!successRulesInGroups.isEmpty) {
          val actualRulesForChildren = new mutable.ListBuffer[Rule]
          actualRulesForChildren ++= actualRules
          actualRulesForChildren --= rulesInGroups

          val targetVariablesInGroup = readyGroups.map(_._1).filter(v => successRulesInGroups.find(
            r => r.collectAssignedVariables.find(v2 => v2.varName.equals(v)).isDefined).isDefined).toSet
          generateNonIntersectingCombinations(successRulesInGroups, targetVariablesInGroup).foreach(c => {
            env.pushContext(context.copy)

            var canBeExecutedChildRules = findCanBeExecutedChildRules(actualRulesForChildren.toList, c)

            val cantBeExecuted = new mutable.HashSet[Rule]
            cantBeExecuted ++= actualRulesForChildren
            cantBeExecuted --= canBeExecutedChildRules

            val child = new FCNode(this, env.currentContext, probabilityContext.copy, canBeExecutedChildRules)
            child.cantBeExecutedRules ++= cantBeExecuted
            children += child

            c.foreach(r => {
              executeConclusions(r)
              child.successRules += r
            })

            env.popContext()
          })

          if (!children.isEmpty) {
            children.foreach(n => n.populate())
            return
          }
        }

        val rulesRequiredUserInput = actualRules.filter(
          r => {
            val notDefined = r.collectAllNotDefinedVariables
            !notDefined.isEmpty && notDefined.forall(v => userVariables.contains(v.varName))
          })
        if (!rulesRequiredUserInput.isEmpty) {
          val targetRule = rulesRequiredUserInput.minBy(_.collectAllNotDefinedVariables.map(v =>
            input.getIndexInDefined(v.varName)).min)
          targetRule.collectAllNotDefinedVariables.foreach(v => {
            if (!v.isDefined) {
              val value = input.getVariableValue(v.varName)
              allContexts.foreach(c => c.set(v.varName, new ProbableValue[Any](value, null)))
            }
          })
        }
        else {
          repeat = false
        }
      }
    }

    private def findCanBeExecutedChildRules(actualRulesForChildren: List[Rule],
                                    childCombination: List[Rule]): List[Rule] = {
      var canBeExecutedChildRules = actualRulesForChildren
      var oldCount = 0
      while (canBeExecutedChildRules.size != oldCount) {
        oldCount = canBeExecutedChildRules.size
        canBeExecutedChildRules =
          canBeExecutedChildRules.filter(r => r.collectAllNotDefinedVariables.map(v => v.varName).forall(vn => {
            userVariables.contains(vn) ||
            childCombination.find(cr => cr.collectAssignedVariables.find(v => v.varName.equals(vn)).isDefined).isDefined ||
            canBeExecutedChildRules.find(r2 => r != r2 && r2.collectAssignedVariables.find(
              v => v.varName.equals(vn)).isDefined).isDefined
          }))
      }
      canBeExecutedChildRules
    }

    private def generateNonIntersectingCombinations(aRules: List[Rule],
                                            targetVariablesInGroups: Set[String]): List[List[Rule]] = {
      val res = new ListBuffer[List[Rule]]
      val currCombination = new mutable.Stack[Rule]
      val assignmentVariablesSet = new mutable.HashSet[String]

      def generateCombinationsInternal(rules: List[Rule]) {
        if (rules.isEmpty) {
          if (!currCombination.isEmpty && targetVariablesInGroups.forall(assignmentVariablesSet.contains(_))) {
            res += currCombination.toList
          }
          return
        }
        else {
          val r = rules.head
          val variables = getRuleAssignedVariables(r)
          if (variables.forall(!assignmentVariablesSet.contains(_))) {
            currCombination.push(r)
            assignmentVariablesSet ++= variables
            generateCombinationsInternal(rules.tail)
            assignmentVariablesSet --= variables
            currCombination.pop()
          }

          generateCombinationsInternal(rules.tail)
        }
      }
      generateCombinationsInternal(aRules)

      res.toList
    }

    private def getRulesGroups: Map[String, List[Rule]] = {
      actualRules.flatMap(r => r.conclusions.map(
        c => (c.variable.varName, r))).groupBy(p => p._1).filter(p => p._2.size > 1).map(
        e => (e._1, e._2.map(p2 => p2._2).toList))
    }

    private def getRulesInGroups(groups: Map[String, List[Rule]]): Set[Rule] = {
      groups.flatMap(e => e._2).toSet
    }

    private def executeRule(r: Rule): Boolean = {
      if (!r.isReadyForExecution) throw new Exception("Attempt to execute not ready rule.")

      var res = false
      if (r.allPreconditionsAreTrue) {
        executeConclusions(r)
        successRules += r
        res = true
      }
      else {
        failedRules += r
      }

      actualRules -= r
      res
    }


    private def executeConclusions(r: Rule) {
      r.executeConclusions(probabilityContext)
    }

    /**
     * @param p - probability of current subtree
     * @param solutions - (OUT parameter) solutions map which is filled after call this method
     */
    def findSolutions(p: Double, solutions: mutable.HashMap[Any, Double]) {
      if (!children.isEmpty) {
        children.foreach(n => n.findSolutions(p/children.size, solutions))
      }
      else if (context.contains(input.goalVariableName)) {
        val probableValue = context.get(input.goalVariableName)
        val value = probableValue.value
        if (!solutions.contains(value)) {
          solutions(value) = p*probableValue.calcProbabilityValue
        }
        else {
          solutions(value) = solutions(value) + p*probableValue.calcProbabilityValue
        }
      }
    }

    private def toStringInternal(indentSize: Int): String = {
      def rulesToString(rules: List[Rule]): String = {
        if (rules.isEmpty)
          "-"
        else {
          val rulesStrings = rules.map(r => r.index.toString)
          rulesStrings.tail.foldLeft(rulesStrings.head)((acc, next) => acc + ", " + next)
        }
      }

      val indent = " "*indentSize

      val LINE_BR = "\n"
      val sb = new mutable.StringBuilder()
      sb.append(indent).append("-- Node --").append(LINE_BR)

      if (!actualRules.isEmpty)
        sb.append(indent).append("Actual rules: " + rulesToString(actualRules.toList.sortBy(_.index))).append(LINE_BR)

      if (!successRules.isEmpty)
        sb.append(indent).append("Success: " + rulesToString(successRules.toList.sortBy(_.index))).append(LINE_BR)

      if (!failedRules.isEmpty)
        sb.append(indent).append("Failed: " + rulesToString(failedRules.toList)).append(LINE_BR)

      if (!cantBeExecutedRules.isEmpty)
        sb.append(indent).append("Can't be executed: " + rulesToString(cantBeExecutedRules.toList)).append(LINE_BR)

      sb.append(indent).append("Context: ").append(context.variableValues).append(LINE_BR)

      if (!children.isEmpty) {
        sb.append(indent).append("Children: ").append(LINE_BR)

        val childIndentSize = indentSize + 3
        children.foreach(n => sb.append(n.toStringInternal(childIndentSize)))
      }

      sb.toString()
    }

    override def toString: String = toStringInternal(0)
  }

  private var env: Environment = null
  private var input: InferenceEngineInput = null
  private val userVariables = new mutable.HashSet[String]
  private val allContexts = new mutable.ListBuffer[Context]
  private var root: FCNode = null
  private val probabilityContext: Context = new Context
  private var allFuzzyExpressions: List[FuzzyBooleanExpression] = null
  private val requiredVariableNameToRulesMap = new mutable.HashMap[String, mutable.ListBuffer[Rule]]

  def getRoot = root

  private def initInput(input: InferenceEngineInput) {
    if (env != null) throw new IllegalStateException("Attempt to run engine twice!")

    env = new Environment
    this.input = input

    userVariables ++= input.variables.map(v => v.varName)
    userVariables --= input.rules.flatMap(r => r.collectAssignedVariables.map(v => v.varName))

    input.variables.foreach(v => v.env = env)
    allFuzzyExpressions = input.rules.flatMap(r => r.preconditions.flatMap(
      p => p.filterExpressions(e => e.isInstanceOf[FuzzyBooleanExpression]))).asInstanceOf[List[FuzzyBooleanExpression]]

    env.contextsStack.push(new Context())
    input.readRequiredVariables()

    var i = 0
    input.rules.foreach(r => {
      r.index = i
      i += 1
      r.collectAllNotDefinedVariables.foreach(v => {
        val name = v.varName
        var rules = requiredVariableNameToRulesMap.getOrElse(name, null)
        if (rules == null) {
          rules = new mutable.ListBuffer[Rule]
          requiredVariableNameToRulesMap(name) = rules
        }
        rules += r
      })
    })
  }

  def run(input: InferenceEngineInput): List[InferenceResult[Any]] = {
    val goal: String = input.goalVariableName
    if (goal == null)
      throw new IllegalArgumentException("gaol variable name is null")

    initInput(input)

    val goalRules = input.rules.filter(r => getRuleAssignedVariables(r).contains(goal))
    if (goalRules.size == 0)
      throw new IllegalArgumentException("gaol variable name is not assigned in rules")

    root = new FCNode(null, env.currentContext, probabilityContext, input.rules)
    root.populate()

    val solutions = new mutable.HashMap[Any, Double]
    root.findSolutions(1.0, solutions)
    solutions.map(p => new InferenceResult[Any](p._1, p._2)).toList
  }

  private def getRuleAssignedVariables(r: Rule): List[String] = {
    r.collectAssignedVariables.map(v => v.varName)
  }
}
