package net.sf.brunneng.fusie.input

import net.sf.brunneng.fusie.engine.VariableValuesSource
import net.sf.brunneng.fusie.core.Variable
import net.sf.brunneng.fusie.input.variables._
import net.sf.brunneng.fusie.ui.UserInteraction
import scala.collection.mutable

/**
 * User: goodg_000
 * Date: 02.05.13
 * Time: 18:23
 */
class StandardValidatedValuesSource(variableDefinitions: List[VariableDefinition[Any]],
                                    ui: UserInteraction) extends VariableValuesSource {

  def read[T](v: Variable[T]): T = {
    val vd = variableDefinitions.find(vd => vd.name.equals(v.varName)).get

    var question = vd.question
    if (question == null) {
      question = "Please enter " + vd.varType + " value of " + vd.name
    }

    var variants: Array[Any] = null
    var variantLabels: List[String] = null
    val pv = vd.possibleValues
    if (pv != null) {
      question += possibleValuesString(pv)

      if (pv.isInstanceOf[EnumPossibleValues[_]]) {
        val epv = pv.asInstanceOf[EnumPossibleValues[Any]]
        variants = epv.getPossibleValues.toArray
        variantLabels = epv.getPossibleValuesLabels
      }
    }

    var parsedValue = null.asInstanceOf[T]
    var valueIsGood = false
    while (!valueIsGood) {
      try {
        val value = ui.readValue(question, variantLabels)
        if (variants != null) {
          parsedValue = variants(variantLabels.indexOf(value)).asInstanceOf[T]
        }
        else {
          parsedValue = parse(v, value)
        }

        if (pv != null) {
          if (!vd.possibleValues.canBeAssigned(parsedValue)) {
            ui.showError("Entered value did not match to restrictions " + possibleValuesString(pv))
          }
          else {
            valueIsGood = true
          }
        }
        else {
          valueIsGood = true
        }
      }
      catch {
        case e:Exception => ui.showError("Can't parse value of type " + v.clazz + ": " + e.getMessage)
      }
    }

    parsedValue
  }

  def parse[T](v: Variable[T], value: String): T = {
    var res:Any = null
    if (v.clazz.equals(classOf[Double])) {
      res = java.lang.Double.parseDouble(value)
    }
    else if (v.clazz.equals(classOf[Int])) {
      res = Integer.parseInt(value)
    }
    else if (v.clazz.equals(classOf[Boolean])) {
      res = java.lang.Boolean.parseBoolean(value)
    }
    else if (v.clazz.equals(classOf[String])) {
      res = value
    }

    res.asInstanceOf[T]
  }

  def possibleValuesString(pv : PossibleValuesDefinition[_]): String = {
    pv match {
      case dr: DoubleRangePossibleValues=> {
        " (" + (if(dr.min!= null) "min: " + dr.min else "") + (if(dr.max!= null) " max: " + dr.max else "") + ")"
      }
      case ir: IntRangePossibleValues=> {
        " (" + (if(ir.min!= null) "min: " + ir.min else "") + (if(ir.max!= null) " max: " + ir.max else "") + ")"
      }
      case er: EnumPossibleValues[_] => {
        var i = 1
        val nextLine = "\n"
        val sb = new mutable.StringBuilder
        sb.append(nextLine)
        er.getPossibleValuesLabels.foreach(v => {
          sb.append("   ").append(i).append(") ").append(v).append(nextLine)
          i += 1
        })
        sb.delete(sb.length - nextLine.size, sb.length)
        sb.toString()
      }
    }
  }

  override def isRequired[T](v: Variable[T]): Boolean = {
    val definitionOpt = variableDefinitions.find(vd => vd.name.equals(v.varName))
    !definitionOpt.isEmpty && definitionOpt.get.required
  }
}
