package net.sf.brunneng.fusie.input.pdl

import scala.util.parsing.combinator._
import net.sf.brunneng.fusie.core.expressions._
import net.sf.brunneng.fusie.core.{Rule, Variable}
import net.sf.brunneng.fusie.core.statements.AssignmentStatement
import net.sf.brunneng.fusie.input.variables._
import scala.collection.mutable._
import scala.Predef._
import scala.util.parsing.input.Position
import scala.collection.mutable
import net.sf.brunneng.fusie.engine.InferenceEngineInput
import net.sf.brunneng.fusie.input.exceptions._
import scala.collection
import net.sf.brunneng.fusie.input._
import net.sf.brunneng.fusie.ui.UserInteraction
import net.sf.brunneng.fusie.input.ast.operations.{TypeUnsafeUnaryOp, TypeUnsafeOp, TypeUnsafeBinOp}
import net.sf.brunneng.fusie.input.ast.functions.{TypeUnsafeFuzzyBooleanFunc, TypeUnsafeFunc}
import net.sf.brunneng.fusie.input.functions.{DefaultFunctionsSource, FunctionsSource}
import net.sf.brunneng.fusie.input.pdl.exceptions.{FunctionCallParsingException, FunctionNotProvidedException}
import net.sf.brunneng.fusie.input.functions.Function
import net.sf.brunneng.fusie.core.expressions.fuzzy._
import net.sf.brunneng.fusie.core.expressions.fuzzy.MembershipFunctionShapePartType
import MembershipFunctionShapePartType._

/**
 * Problem Definition Language parser
 *
 *
 * User: goodg_000
 * Date: 13.04.13
 * Time: 20:43
 */
class PDLParser extends JavaTokenParsers {

  protected val positions = new mutable.HashMap[Any, Position]
  protected val textRepresentation = new mutable.HashMap[Any, String]
  protected var pd: ProblemDefinition = null
  protected var functionsSource: FunctionsSource = null

  def createProblemDefinition(input: String): ProblemDefinition = {
    val parseResult = parseAll(problemDefinition, input)
    if (parseResult.isEmpty) {
      throw new ParsingException(parseResult.toString)
    }

    pd = parseResult.get
    pd
  }

  def validateProblemDefinitionAndMakeInput(functionsSource: FunctionsSource = null,
                                            ui: UserInteraction = null): InferenceEngineInput = {
    this.functionsSource = if (functionsSource != null) functionsSource else new DefaultFunctionsSource()

    validateUndefinedVariables(pd)
    validateNoConflictedVariableDefinitions(pd)
    validateNoTypeConflictWithDefinedVariables(pd)
    validateNoTypeConflictsForVariables(pd)
    validateNoDefinedVariablesInConclusions(pd)
    inferTypesOfConclusions()
    inferTypesOfPreconditionsAndProbability()

    new InferenceEngineInput(pd.variableDefinitions.map(vd => vd.name), pd.getAllVariablesWithAssigned,
      pd.rules, pd.target, new StandardValidatedValuesSource(pd.variableDefinitions, ui))
  }

  def parseProblem(input: String, functionsSource: FunctionsSource = null,
                   ui: UserInteraction = null) : InferenceEngineInput = {
    createProblemDefinition(input)
    validateProblemDefinitionAndMakeInput(functionsSource, ui)
  }

  protected def inferTypesOfConclusions() {
    val definitionsMap = pd.getDefinitionsMap
    pd.getVariablesWithUndefinedTypes.foreach(v => {
      if (definitionsMap.contains(v.varName)) {
        v.clazz = definitionsMap(v.varName).varType
      }
    })

    var end = false
    while (!end) {
      val conclusionsWithUndefinedType = pd.getConclusionsWithUndefinedTypes
      if (!conclusionsWithUndefinedType.isEmpty) {
        var inferedTypeForAtLeastOne = false
        conclusionsWithUndefinedType.foreach(as => {
          if (as.expr.getAllVariables.forall(pd.variableWithDefinedTypeP)) {
            inferTypes(as)
            setType(as.variable.varName, as.variable.clazz)
            validateNoTypeConflictsForVariables(pd)
            inferedTypeForAtLeastOne = true
          }
        })

        if (!inferedTypeForAtLeastOne) {
          val v = conclusionsWithUndefinedType.map(as => as.variable.varName).toSet
          throw new UnableToInferTypesForAssignementsException(v, "Unable to infer types for assignments:\n" +
             objectsToString(conclusionsWithUndefinedType))
        }
      }
      else {
        end = true
      }
    }

    pd.getConclusionsWithUndefinedExpressions.foreach(as => inferTypes(as))
  }

  protected def inferTypesOfPreconditionsAndProbability() {
    pd.rules = pd.rules.map(r => inferTypesOfPreconditionsAndProbability(r))
  }

  protected def inferTypesOfPreconditionsAndProbability(r: Rule): Rule = {
    val processedPreconditions = r.preconditions.map(e => {
      val p = inferTypes(e)
      if (!p._1.equals(classOf[Boolean])) {
        throw new TypeMismatchException("Expected type of " + objMetaToString(e) +
          " is boolean, but actually is " + p._1)
      }
      p._2.asInstanceOf[Expression[Boolean]]
    })

    var processedProbability:Expression[Double] = null
    val p = inferTypes(r.probability)
    if (!p._1.equals(classOf[Double])) {
      throw new TypeMismatchException("Expected type of " + objMetaToString(r.probability) +
        " is double, but actually is " + p._1)
    }
    else {
      processedProbability = p._2.asInstanceOf[Expression[Double]]
    }

    new Rule(processedPreconditions, r.conclusions, processedProbability)
  }

  protected def setType(varName: String, clazz: Class[Any]) {
    pd.getAllVariablesWithAssigned.filter(pd.variableWithUndefinedTypeP).filter(_.varName.equals(varName)).foreach(
      v => v.clazz = clazz)
  }

  protected def inferTypes(as : AssignmentStatement[Any]) {
    val v = as.variable

    val p = inferTypes(as.expr)
    var inferredExpr = p._2
    if (pd.variableWithDefinedTypeP(v)) {
      if (!v.clazz.equals(p._1)) {
        val castedExpression = tryCast(inferredExpr, p._1, v.clazz)
        if (castedExpression != null) {
          inferredExpr = castedExpression
        }
        else {
          throw new AssignmentTypeCastException(v.varName, "Unable to assign expression " +
            objMetaToString(as.expr) + " of type " + p._1 + " to " + objMetaToString(v) +
            " of type " + v.clazz)
        }
      }
    }

    v.clazz = p._1.asInstanceOf[Class[Any]]
    as.expr = inferredExpr
  }

  protected def cast(expr: Expression[_],
                   fromClass: Class[_], toClass: Class[_]): Expression[_] = {
    val res = tryCast(expr, fromClass, toClass)
    if (res == null) {
      throw new TypeMismatchException("Unable to cast expression " + objMetaToString(expr) +
        " from " + fromClass + " to " + toClass)
    }
    res
  }

  protected def tryCast(expr: Expression[_],
                      fromClass: Class[_], toClass: Class[_]): Expression[_] = {
    if (fromClass.equals(classOf[Int]) && toClass.equals(classOf[Double]))
      F1ExpressionFactory.castToDouble(expr.asInstanceOf[Expression[Int]])
    else if (fromClass.equals(toClass))
      expr
    else
      null
  }

  protected def inferTypes(e : Expression[Any]): (Class[_], Expression[_]) = {
    def toPrimitive(c: Class[_]): Class[_] = {
      if (c.equals(classOf[java.lang.Boolean]))
        classOf[Boolean]
      else if (c.equals(classOf[java.lang.Integer]))
        classOf[Int]
      else if (c.equals(classOf[java.lang.Double]))
        classOf[Double]
      else
        c
    }

    val childExpressions = e.getChildExpressions
    var res: (Class[_], Expression[_]) = null
    if (childExpressions.isEmpty) {
      res = e match {
        case ve: ValueExpression[Any] => (toPrimitive(ve.value.getClass).asInstanceOf[Class[Any]], e)
        case ve: VariableExpression[Any] => (ve.variable.clazz, e)
        case _ => throw new InternalParsingError("Unknown type of expression " + objMetaToString(e))
      }
    }
    else {
      if (e.isInstanceOf[TypeUnsafeOp]) {
        res = convertToTypeInferredExpression(
          e.asInstanceOf[TypeUnsafeOp], childExpressions.map(inferTypes(_)).toArray)
      }
      else if (e.isInstanceOf[TypeUnsafeFunc[_]]) {
        if (functionsSource == null) {
          throw new ParsingException("Can't infer types of " + objMetaToString(e) +
            " because functions source is not specified ")
        }

        res = convertFuncToTypeInferredExpression(
          e.asInstanceOf[TypeUnsafeFunc[_]], childExpressions.map(inferTypes(_)).toArray)
      }
      else if (e.isInstanceOf[TypeUnsafeFuzzyBooleanFunc]) {
        res = convertFuzzyFuncToTypeInferredExpression(
          e.asInstanceOf[TypeUnsafeFuzzyBooleanFunc], childExpressions.map(inferTypes(_)).toArray)
      }
      else {
        throw new InternalParsingError("Can't define type of expression " + objMetaToString(e))
      }

      positions(res._2) = positions(e)
      textRepresentation(res._2) = textRepresentation(e)
    }

    res
  }

  protected def convertFuncToTypeInferredExpression(func: TypeUnsafeFunc[_],
      childSignatures : Array[(Class[_], Expression[_])]): (Class[_], Expression[_]) = {

    val functions = functionsSource.getFunctions(func.name)
    if (functions.isEmpty) {
      throw new FunctionNotProvidedException(func.name)
    }

    var incorrectNumberOfArgumentsDetected = true

    var maxTypeMismatchExceptionParamNum = -1
    var typeMismatchException: TypeMismatchException = null

    var matchedFunction: Function = null
    val castedExpressions = new ListBuffer[Expression[_]]

    functions.find(f => {
      val m = f.method
      val parameterTypes = m.getParameterTypes
      if (parameterTypes.size == childSignatures.size) {
        incorrectNumberOfArgumentsDetected = false

        val argsSize = parameterTypes.size
        var i = 0

        try {
          castedExpressions.clear()
          while (i < argsSize) {
            val expectedType = parameterTypes(i)
            val expr = childSignatures(i)

            castedExpressions += cast(expr._2, expr._1, expectedType)

            i += 1
          }
          matchedFunction = f
        }
        catch {
          case e:TypeMismatchException => {
            if (i > maxTypeMismatchExceptionParamNum) {
              typeMismatchException = e
              maxTypeMismatchExceptionParamNum = i
            }
          }
        }
      }

      matchedFunction != null
    })

    if (incorrectNumberOfArgumentsDetected) {
      throw new FunctionCallParsingException(func.name, "Invalid number of arguments to function call " +
        objMetaToString(func))
    }

    if (matchedFunction == null) {
      throw typeMismatchException
    }

    (matchedFunction.method.getReturnType, new FnReflectionExpression[Any](func.name, castedExpressions.toList,
      matchedFunction.obj, matchedFunction.method))
  }

  protected def convertFuzzyFuncToTypeInferredExpression(func: TypeUnsafeFuzzyBooleanFunc,
      childSignatures : Array[(Class[_], Expression[_])]): (Class[_], Expression[_]) = {

    val target = childSignatures.head

    val doubleTarget = cast(target._2, target._1, classOf[Double]).asInstanceOf[Expression[Double]]
    val doubleArguments = childSignatures.tail.map(
      p => cast(p._2, p._1, classOf[Double])).toList.asInstanceOf[List[Expression[Double]]]

    val shape = new ListBuffer[MembershipFunctionShapePartType]
    func.fuzzyFuncShape.foreach(ch => {
      shape += (ch match {
        case '/' => MembershipFunctionShapePartType.GROWTH
        case '\\' => MembershipFunctionShapePartType.DECREASE
      })})

    if (doubleArguments.size != shape.size + 1) {
      throw new ParsingException("In fuzzy function " + objMetaToString(func) +
        " count of points (actual " + doubleArguments.size + ") should be 1 more then count of shape parts (actual " +
        shape.size + ")" )
    }

    val membershipProbabilitySource = new MembershipFunctionProbabilitySource(shape.toList)

    func.name match {
      case "linear" => {
        (classOf[Boolean],
          new FuzzyBooleanExpression(
            new LinearMembershipFunction(doubleTarget, membershipProbabilitySource, doubleArguments)))
      }
      case "poly" => {
        (classOf[Boolean],
          new FuzzyBooleanExpression(
            new PolynomialMembershipFunction(doubleTarget, membershipProbabilitySource, doubleArguments)))
      }
      case other => {
        throw new ParsingException("Fuzzy membership function '" + other + "' not supported")
      }
    }
  }

  private def convertToTypeInferredExpression(op: TypeUnsafeOp,
       childSignatures : Array[(Class[_], Expression[_])]): (Class[_], Expression[_]) = {
    def to[T](e : Expression[_]): Expression[T] = e.asInstanceOf[Expression[T]]

    try {
      op.getOperation match {
        case "+" => {
          val s = twoIntOrTwoDouble(childSignatures)
          if (s._1.equals(classOf[Int]))
            (classOf[Int], F2ExpressionFactory.plus(to[Int](s._2), to[Int](s._3)))
          else
            (classOf[Double], F2ExpressionFactory.plus(to[Double](s._2), to[Double](s._3)))
        }
        case "-" => {
          if (op.isInstanceOf[TypeUnsafeUnaryOp[_, _]]) {
            validateSignatureSize(childSignatures, 1)
            val inner = childSignatures.head
            if (inner._1.equals(classOf[Int])) {
              (classOf[Int], F1ExpressionFactory.negate(inner._2.asInstanceOf[Expression[Int]]))
            }
            else if (inner._1.equals(classOf[Double])) {
              (classOf[Double], F1ExpressionFactory.negate(inner._2.asInstanceOf[Expression[Double]]))
            }
            else {
              throw new TypeMismatchException("Expected type of " + objMetaToString(inner._2) + " should be numeric")
            }
          }
          else {
            val s = twoIntOrTwoDouble(childSignatures)
            if (s._1.equals(classOf[Int]))
              (classOf[Int], F2ExpressionFactory.minus(to[Int](s._2), to[Int](s._3)))
            else
              (classOf[Double], F2ExpressionFactory.minus(to[Double](s._2), to[Double](s._3)))
          }
        }
        case "*" => {
          val s = twoIntOrTwoDouble(childSignatures)
          if (s._1.equals(classOf[Int]))
            (classOf[Int], F2ExpressionFactory.multiply(to[Int](s._2), to[Int](s._3)))
          else
            (classOf[Double], F2ExpressionFactory.multiply(to[Double](s._2), to[Double](s._3)))
        }
        case "/" => {
          val s = twoIntOrTwoDouble(childSignatures)
          if (s._1.equals(classOf[Int]))
            (classOf[Int], F2ExpressionFactory.integralDiv(to[Int](s._2), to[Int](s._3)))
          else
            (classOf[Double], F2ExpressionFactory.div(to[Double](s._2), to[Double](s._3)))
        }
        case "^" => {
          val s = twoIntOrTwoDouble(childSignatures)
          if (s._1.equals(classOf[Int]))
            (classOf[Int], F2ExpressionFactory.pow(to[Int](s._2), to[Int](s._3)))
          else
            (classOf[Double], F2ExpressionFactory.pow(to[Double](s._2), to[Double](s._3)))
        }
        case "==" => {
          val s = twoSameClass(childSignatures)
          if (s._1.equals(classOf[Double]))
            (classOf[Boolean], F2ExpressionFactory.eqDouble(
              s._2.asInstanceOf[Expression[Double]], s._3.asInstanceOf[Expression[Double]]))
          else
            (classOf[Boolean], F2ExpressionFactory.eq(to[Any](s._2), to[Any](s._3)))
        }
        case "!=" => {
          val s = twoSameClass(childSignatures)
          if (s._1.equals(classOf[Double]))
            (classOf[Boolean], F2ExpressionFactory.notEqDouble(
              s._2.asInstanceOf[Expression[Double]], s._3.asInstanceOf[Expression[Double]]))
          else
            (classOf[Boolean], F2ExpressionFactory.notEq(to[Any](s._2), to[Any](s._3)))
        }
        case "<" => {
          val s = twoIntOrTwoDouble(childSignatures)
          if (s._1.equals(classOf[Int]))
            (classOf[Boolean], F2ExpressionFactory.lt(to[Int](s._2), to[Int](s._3)))
          else
            (classOf[Boolean], F2ExpressionFactory.lt(to[Double](s._2), to[Double](s._3)))
        }
        case ">" => {
          val s = twoIntOrTwoDouble(childSignatures)
          if (s._1.equals(classOf[Int]))
            (classOf[Boolean], F2ExpressionFactory.gt(to[Int](s._2), to[Int](s._3)))
          else
            (classOf[Boolean], F2ExpressionFactory.gt(to[Double](s._2), to[Double](s._3)))
        }
        case "<=" => {
          val s = twoIntOrTwoDouble(childSignatures)
          if (s._1.equals(classOf[Int]))
            (classOf[Boolean], F2ExpressionFactory.lteq(to[Int](s._2), to[Int](s._3)))
          else
            (classOf[Boolean], F2ExpressionFactory.lteq(to[Double](s._2), to[Double](s._3)))
        }
        case ">=" => {
          val s = twoIntOrTwoDouble(childSignatures)
          if (s._1.equals(classOf[Int]))
            (classOf[Boolean], F2ExpressionFactory.gteq(to[Int](s._2), to[Int](s._3)))
          else
            (classOf[Boolean], F2ExpressionFactory.gteq(to[Double](s._2), to[Double](s._3)))
        }
        case "&&" => {
          val s = twoSameClass(childSignatures, collection.Set(classOf[Boolean]))
          (s._1, F2ExpressionFactory.and(to[Boolean](s._2), to[Boolean](s._3)))
        }
        case "||" => {
          val s = twoSameClass(childSignatures, collection.Set(classOf[Boolean]))
          (s._1, F2ExpressionFactory.or(to[Boolean](s._2), to[Boolean](s._3)))
        }
        case "!" => {
          validateSignatureSize(childSignatures, 1)
          val s = childSignatures(0)
          (s._1, F1ExpressionFactory.not(to[Boolean](s._2)))
        }
        case _ => throw new InternalParsingError("Unknown operation " + op)
      }
    }
    catch {
      case e:TypeMismatchException => throw new TypeMismatchException("Type error while processing operation " +
        objMetaToString(op) + ": " + e.getMessage)
    }
  }

  protected def twoIntOrTwoDouble(signatures: Array[(Class[_], Expression[_])]):
    (Class[_], Expression[_], Expression[_]) = {

    twoSameClass(signatures, collection.Set(classOf[Int], classOf[Double]))
  }

  protected def twoSameClass(signatures: Array[(Class[_], Expression[_])],
                           classes: collection.Set[Class[_]] =
                              collection.Set(classOf[Int], classOf[Double], classOf[Boolean], classOf[String])):
    (Class[_], Expression[_], Expression[_]) = {
    validateSignatureSize(signatures, 2)

    val c1 = signatures(0)._1
    val e1 = signatures(0)._2
    val c2 = signatures(1)._1
    val e2 = signatures(1)._2

    val doubleClass = classOf[Double]
    if (c1.equals(c2) && classes.contains(c1)) {
      (c1, e1, e2)
    }
    else if (classes.contains(doubleClass)) {
      val castedExpr1 = cast(e1, c1, doubleClass)
      val castedExpr2 = cast(e2, c2, doubleClass)

      (doubleClass, castedExpr1, castedExpr2)
    }
    else {
      throw new TypeMismatchException("Both types of expressions " + objectsToString(signatures.map(_._2).toList, ", ")
        + " should be one of " + classes)
    }
  }

  protected def validateSignatureSize(signatures: Array[(Class[_], Expression[_])], expectedSize: Int) {
    if (signatures.size != expectedSize) {
      throw new InternalParsingError("Expected signatures size is " + expectedSize + " but was " + signatures.size)
    }
  }

  protected def validateUndefinedVariables(pd: ProblemDefinition) {
    val allVariables = pd.getAllVariables.map(v => v.varName)

    val definedVarNames = new mutable.HashSet[String]()
    pd.variableDefinitions.foreach(vd => definedVarNames += vd.name)
    pd.rules.foreach(r => definedVarNames ++= r.collectAssignedVariables.map(v => v.varName))

    val notDefinedVarNames = allVariables.filter(v => !definedVarNames.contains(v))
    if (!notDefinedVarNames.isEmpty) {
      val notDefinedVariables = new mutable.ListBuffer[Variable[Any]]()
      pd.rules.foreach(r => {
        notDefinedVariables ++= r.collectAllVariables.filter(
          v => notDefinedVarNames.contains(v.varName))
      })

      throw new NotDefinedVariablesException(notDefinedVarNames.toSet,
        "Variables not defined: \n" + objectsToString(notDefinedVariables.toList))
    }
  }

  protected def validateNoConflictedVariableDefinitions(pd: ProblemDefinition) {
    val varNames = new mutable.HashSet[String]
    val conflictedDefinitions = new ListBuffer[VariableDefinition[Any]]
    pd.variableDefinitions.foreach(vd => {
      if (varNames.contains(vd.name)) {
        conflictedDefinitions += vd
      }
      else {
        varNames += vd.name
      }
    })

    if (!conflictedDefinitions.isEmpty) {
      throw new AlreadyDefinedVariablesException(conflictedDefinitions.map(vd => vd.name).toSet,
        "Variables already defined: \n" + objectsToString(conflictedDefinitions.toList))
    }
  }

  protected def validateNoTypeConflictWithDefinedVariables(pd: ProblemDefinition) {
    val varDefinitionsMap = pd.getDefinitionsMap

    var conflictedVariables = new mutable.ListBuffer[Variable[Any]]
    pd.rules.foreach(r => {
      conflictedVariables ++= r.collectAllVariables.filter(v => varDefinitionsMap.contains(v.varName) &&
        pd.variableWithDefinedTypeP(v) && !v.clazz.isAssignableFrom(varDefinitionsMap(v.varName).varType))
    })

    if (!conflictedVariables.isEmpty) {
      val sb = new mutable.StringBuilder()
      conflictedVariables.foreach(v => {
        sb.append(objMetaToString(v) + " expected type is " + v.clazz + " but defined type is " +
          varDefinitionsMap(v.varName).varType + "\n")
      })

      throw new VariablesTypeConflictWithDefinitionException(conflictedVariables.map(v => v.varName).toSet,
        "Variable type conflicts with definitions:\n" + sb)
    }
  }

  protected def validateNoTypeConflictsForVariables(pd: ProblemDefinition) {
    val conflictedVariablesSet = new mutable.HashSet[String]()

    val sb = new mutable.StringBuilder()
    pd.getAllVariablesWithAssigned.filter(pd.variableWithDefinedTypeP).groupBy(_.varName).foreach(p => {
      val classesGroups = p._2.groupBy(_.clazz)
      if (classesGroups.size > 1) {
        conflictedVariablesSet += p._1
        sb.append("Variable " + p._1 + ": ")
        sb.append(joinStrings(classesGroups.map(cg => objectsPositionsToString(cg._2) +
          " expected type is " + cg._1).toList, "; "))
        sb.append("\n")
      }
    })

    if (!conflictedVariablesSet.isEmpty) {
      throw new VariableTypeConflictsException(conflictedVariablesSet.toSet,
        "Detected conflicts of variable types:\n" + sb.toString())
    }
  }

  protected def validateNoDefinedVariablesInConclusions(pd: ProblemDefinition) {
    val definitionsMap = pd.getDefinitionsMap
    val conclusions = pd.getAllConclusions

    val conclusionsInDefinitions = conclusions.filter(as => definitionsMap.contains(as.variable.varName))
    if (!conclusionsInDefinitions.isEmpty) {
      throw new DefinedVariablesInConclusionsException(conclusionsInDefinitions.map(as => as.variable.varName).toSet,
         "Detected assignments to defined user variables:\n" + objectsToString(conclusionsInDefinitions))
    }
  }

  protected def objectsToString(objects : List[Any], separator: String = "\n") : String = {
    joinStrings(objects.map(o => objMetaToString(o)), separator)
  }

  protected def objectsPositionsToString(objects : List[Any]) : String = {
    joinStrings(objects.map(o => objPositionToString(o)), ", ")
  }

  protected def objMetaToString(obj : Any) : String = {
    if (!textRepresentation.contains(obj)) {
      return "Provided object: " + obj;
    }

    val txt = textRepresentation(obj)
    objPositionToString(obj) + " " + txt
  }

  protected def objPositionToString(obj: Any): String = {
    val pos = positions(obj)
    "[line: " + pos.line + ", col:" + pos.column + "]"
  }

  protected def joinStrings(strings: List[String], delimiter: String): String = {
    strings.tail.foldLeft(strings.head)((sum, s) => sum + delimiter + s)
  }

  /**
   * Saves metadata of successfully parsed object, such as position in text and text representation
   */
  protected def saveMeta[T](p: => Parser[T]): Parser[T] = Parser { in =>
    p(in) match {
      case Success(t, in1) => {
        positions(t) = in.pos
        textRepresentation(t) = in.source.subSequence(in.offset, in1.offset).toString.trim
        val v = t match {
          case ve: VariableExpression[_] => ve.variable
          case as: AssignmentStatement[_] => as.variable
          case _ => null
        }
        if (v != null) {
          positions(v) = positions(t)
          textRepresentation(v) = textRepresentation(t)
        }

        Success(t, in1)
      }
      case ns: NoSuccess => ns
    }
  }

  protected def saveMeta[T](expr: Expression[T]): Expression[T] = {
    if (expr.isInstanceOf[TypeUnsafeBinOp[_, _, _]]) {
      val binaryOp = expr.asInstanceOf[TypeUnsafeBinOp[_, _, _]]
      val first = binaryOp.getChildExpressions.head
      val second = binaryOp.getChildExpressions.tail.head

      positions(expr) = positions(first)
      textRepresentation(expr) = textRepresentation(first) + binaryOp.getOperation + textRepresentation(second)
    }

    expr
  }

  protected def fold(e: Expression[Any], l: List[~[String,Expression[Any]]]) =
    l.foldLeft(e){ case (e1,op~e2) => saveMeta(new TypeUnsafeBinOp(e1, e2, op))}
  protected def foldBool(e: Expression[Boolean], l: List[~[String,Expression[Boolean]]]) =
      l.foldLeft(e){ case (e1,op~e2) => op match {
        case "&&" => saveMeta(new TypeUnsafeBinOp[Boolean, Boolean, Boolean](e1, e2, "&&"))
        case "||" => saveMeta(new TypeUnsafeBinOp[Boolean, Boolean, Boolean](e1, e2, "||"))
      }}

  protected def toAnyVariable[T](p: Parser[(Class[T], String, PossibleValuesDefinition[T])]):
      Parser[(Class[Any], String, PossibleValuesDefinition[Any])] =
    p.asInstanceOf[Parser[(Class[Any], String, PossibleValuesDefinition[Any])]]

  protected def problemDefinition: Parser[ProblemDefinition] =
    saveMeta((opt(eols)~>opt(variableDefinitions)<~opt(eols))~rules~(kwFind~>identifier)<~opt(eols) ^^
    {case vdOpt~r~i => new ProblemDefinition(vdOpt.getOrElse(List()), r, i)})

  protected def variableDefinitions: Parser[List[VariableDefinition[Any]]] = saveMeta(rep1(variableDefinition<~eols))

  protected def variableDefinition : Parser[VariableDefinition[Any]] = saveMeta(opt(kwRequired)~
    (toAnyVariable(doubleVariable) |
     toAnyVariable(intVariable) |
     toAnyVariable(boolVariable) |
     toAnyVariable(enumVariable)
    )~ opt("<-"~>stringConstant) ^^
    {case rOpt~v~qOpt =>
      new VariableDefinition[Any](rOpt.isDefined, v._1, v._2, v._3, qOpt.getOrElse(null))})

  protected def enumVariable: Parser[(Class[String], String, PossibleValuesDefinition[String])] =
    saveMeta(kwEnum~>identifier~("="~>enumEnum) ^^ {case b~pv => (classOf[String], b, pv)})

  protected def enumEnum: Parser[PossibleValuesDefinition[String]] = saveMeta(rep1sep(enumPossibleValue, "|") ^^
    {case c => new EnumPossibleValues(c)})

  protected def enumPossibleValue: Parser[PossibleValue[String]] = saveMeta(identifier~opt("("~>stringConstant<~")") ^^
    {case i~labelOpt => new PossibleValue(i, labelOpt.getOrElse(i.toString))})

  protected def boolVariable: Parser[(Class[Boolean], String, PossibleValuesDefinition[Boolean])] =
      kwBool~>identifier~opt("="~>boolEnum) ^^
      {case b~pvOpt => (classOf[Boolean], b,
        pvOpt.getOrElse(new EnumPossibleValues(List(new PossibleValue(true, "Yes"), new PossibleValue(false, "No")))))}

  protected def boolEnum: Parser[PossibleValuesDefinition[Boolean]] = saveMeta((stringConstant<~"|")~stringConstant ^^
    {case trueLabel~falseLabel =>
      new EnumPossibleValues(List(new PossibleValue(true, trueLabel), new PossibleValue(false, falseLabel)))})

  protected def intVariable: Parser[(Class[Int], String, PossibleValuesDefinition[Int])] = saveMeta(
    kwInt~>identifier~opt("="~>(intRange | intEnum)) ^^ {case i~pvOpt => (classOf[Int], i, pvOpt.getOrElse(null))})

  protected  def intRange: Parser[PossibleValuesDefinition[Int]] = saveMeta(
    (kwMin~>intNumber)~(kwMax~>intNumber) ^^ {case min~max => new IntRangePossibleValues(min, max)} |
    (kwMin~>intNumber) ^^ {case min => new IntRangePossibleValues(min, null)} |
    (kwMax~>intNumber) ^^ {case max => new IntRangePossibleValues(null, max)})

  protected def intEnum: Parser[PossibleValuesDefinition[Int]] = saveMeta(rep1sep(intPossibleValue, "|") ^^
    {case i => new EnumPossibleValues(i)})

  protected def intPossibleValue: Parser[PossibleValue[Int]] = saveMeta(intNumber~opt("("~>stringConstant<~")") ^^
    {case i~labelOpt => new PossibleValue(i, labelOpt.getOrElse(i.toString))})

  protected def doubleVariable: Parser[(Class[Double], String, PossibleValuesDefinition[Double])] =
    kwDouble~>identifier~opt("="~>(doubleRange | doubleEnum)) ^^
      {case i~pvOpt => (classOf[Double], i, pvOpt.getOrElse(null))}

  protected def doubleRange: Parser[PossibleValuesDefinition[Double]] = saveMeta(
    (kwMin~>doubleNumber)~(kwMax~>doubleNumber) ^^ {case min~max => new DoubleRangePossibleValues(min, max)} |
    (kwMin~>doubleNumber) ^^ {case min => new DoubleRangePossibleValues(min, null)} |
    (kwMax~>doubleNumber) ^^ {case max => new DoubleRangePossibleValues(null, max)})

  protected def doubleEnum: Parser[PossibleValuesDefinition[Double]] = saveMeta(rep1sep(doublePossibleValue, "|") ^^
    {case c => new EnumPossibleValues(c)})

  protected def doublePossibleValue: Parser[PossibleValue[Double]] = saveMeta(
    doubleNumber~opt("("~>stringConstant<~")") ^^
      {case d~labelOpt => new PossibleValue(d, labelOpt.getOrElse(d.toString))})

  protected def rules: Parser[List[Rule]] = saveMeta(rep1((fact | rule)<~eols))

  protected def fact: Parser[Rule] = saveMeta(
      (kwFact~>opt(identifier)<~eol)~
      (assignments<~eol)~
      opt((kwProb~eol)~>expr<~eol) ^^
    {case optRule~a~optProb =>new Rule(
      optRule.getOrElse(null),
      List(new ValueExpression[Boolean](true)), a.asInstanceOf[List[AssignmentStatement[Any]]],
      optProb.getOrElse(new ValueExpression(1.0)).asInstanceOf[Expression[Double]])})

  protected def rule: Parser[Rule] = saveMeta(
      opt(kwRule~>identifier<~eol)~
      ((kwWhen~eol)~>boolExpressions<~eol)~
      ((kwThen~eol)~>assignments<~eol)~
      opt((kwProb~eol)~>expr<~eol) ^^
    {case optRule~be~a~optProb =>
      new Rule(
        optRule.getOrElse(null),
        be, a.asInstanceOf[List[AssignmentStatement[Any]]],
        optProb.getOrElse(new ValueExpression(1.0)).asInstanceOf[Expression[Double]])})

  protected def assignments: Parser[List[AssignmentStatement[Any]]] = saveMeta(rep1sep(assignment, eol))
  protected def boolExpressions: Parser[List[Expression[Boolean]]] = saveMeta(rep1sep(boolExpr, eol))

  protected def assignment: Parser[AssignmentStatement[Any]] = saveMeta(
    identifier~("="~>boolExpr) ^^ {case i~e => new AssignmentStatement(
      new Variable(i, classOf[Any]), e)} |||
    identifier~("="~>expr) ^^ {case i~e => new AssignmentStatement(new Variable(i, classOf[Any]), e)})

  protected def boolExpr: Parser[Expression[Boolean]] = saveMeta(
    boolTerm~rep(("&&" | "||")~boolTerm) ^^ { case b~l => foldBool(b, l) })

  protected def boolTerm: Parser[Expression[Boolean]] = saveMeta(
    expr~(">=" | "<=" | ">" | "<" | "==" | "!=")~expr ^^
      { case e1~op~e2 => new TypeUnsafeBinOp(e1,e2,op).asInstanceOf[Expression[Boolean]] } |
    "("~>boolExpr<~")" |
    fuzzyBoolTerm |
    boolValue)

  protected def fuzzyBoolTerm: Parser[Expression[Boolean]] =
    (expr<~"~=")~(identifier<~"(")~(fuzzyCurveShape<~",")~(repsep(expr, ",")<~")") ^^
      { case e1~op~fcs~args => new TypeUnsafeFuzzyBooleanFunc(op, fcs,
        e1.asInstanceOf[Expression[Double]], args.asInstanceOf[List[Expression[Double]]])}

  protected def fuzzyCurveShape: Parser[String] = """[\\/]+""".r

  protected def boolValue: Parser[Expression[Boolean]] = saveMeta(
    kwTrue ^^ (x => new ValueExpression(true)) |
    kwFalse ^^ (x => new ValueExpression(false)) |
    functionCall.asInstanceOf[Parser[Expression[Boolean]]] |
    identifier ^^ {case i => new VariableExpression(new Variable(i, classOf[Any])).asInstanceOf[Expression[Boolean]]} |
    "!"~>"("~>boolExpr<~")" ^^ (b => new TypeUnsafeUnaryOp[Boolean, Boolean](b, "!")) |
    "("~>boolExpr<~")" |
    "!"~>boolValue ^^ (b => new TypeUnsafeUnaryOp[Boolean, Boolean](b, "!")))

  protected def expr: Parser[Expression[Any]] = saveMeta(term~rep(("+"|"-")~term) ^^ { case t~l => fold(t, l) })

  protected def term: Parser[Expression[Any]] = saveMeta(negFactor~rep(("*"|"/"|"%")~negFactor) ^^ { case f~l => fold(f, l) })

  protected def negFactor: Parser[Expression[Any]] = saveMeta(
    opt("-")~factor ^^ {case optNeg~f => if (optNeg.isDefined) new TypeUnsafeUnaryOp(f, "-") else f}
  )

  protected def factor: Parser[Expression[Any]] = saveMeta(
      doubleNumber ^^ {case d => (new ValueExpression(d))}  |
      intNumber ^^ {case i => new ValueExpression(i)} |
      functionCall |
      identifier ^^ {case i => new VariableExpression(new Variable(i, classOf[Any]))} |
      stringConstant ^^ {case s => new ValueExpression(s)} |
      "("~>expr<~")")

  protected def functionCall: Parser[Expression[Any]] = saveMeta(identifier ~ ("("~>repsep(boolExpr ||| expr, ",")<~")") ^^
    {case name~args => new TypeUnsafeFunc[Any](name, args)})

  protected val stringConstant : Parser[String] = stringLiteral ^^ {case s => s.substring(1, s.length - 1)}
  protected var intNumber: Parser[Int] = wholeNumber ^^ {case i => i.toInt}
  protected val doubleNumber: Parser[Double] = """\d+\.\d*""".r ^^ {case d => d.toDouble}

  protected var identifier: Parser[String] = not(reserved) ~> identWithDots

  protected var identWithDots: Parser[String] = """[a-zA-Z_][\w.]*""".r

  protected val kwMin: Parser[String] = "min:".r
  protected val kwMax: Parser[String] = "max:".r
  protected val kwFind: Parser[String] = "find\\b".r
  protected val kwRequired: Parser[String] = "required\\b".r
  protected val kwEnum: Parser[String] = "enum\\b".r
  protected val kwBool: Parser[String] = "bool\\b".r
  protected val kwInt: Parser[String] = "int\\b".r
  protected val kwDouble: Parser[String] = "double\\b".r
  protected val kwTrue: Parser[String] = "true\\b".r
  protected val kwFalse: Parser[String] = "false\\b".r
  protected val kwFact: Parser[String] = "fact\\b".r
  protected val kwRule: Parser[String] = "rule\\b".r
  protected val kwWhen: Parser[String] = "when\\b".r
  protected val kwThen: Parser[String] = "then\\b".r
  protected val kwProb: Parser[String] = "prob\\b".r
  protected val reserved: Parser[String] = kwFind | kwRequired | kwEnum | kwBool | kwInt | kwDouble |
    kwTrue | kwFalse | kwRule | kwWhen | kwThen | kwProb

  protected def eols: Parser[String] = eol<~rep(eol|whiteSpace)
  protected val eol: Parser[String] = """(\r?\n)""".r
  protected override val whiteSpace = """([ \t]|(?m)/\*(\*(?!/)|[^*])*\*/)+""".r
}
